package sugartracer;

import java.util.ArrayList;
import java.util.List;

import sugartracer.decoder.Decoder;
import sugartracer.decoder.Variable;

/**
 * This is a class for tracer.
 * @author Naoyuki Tamura
 */
public abstract class Tracer {
    public SatSolver solver;
    public Decoder decoder;
    public List<Integer> propagations = null;
    public List<Integer> undos = null;
    public boolean done = false;
    public boolean showAuxVars = false;
    public boolean quick = false;
    public boolean debug = false;
    
    /**
     * Creates the tracer.
     * @param decoder
     */
    public Tracer(SatSolver solver, Decoder decoder) {
        super();
        this.solver = solver;
        this.decoder = decoder;
    }

    /**
     * Shows the strings
     * @param strs
     */
    public abstract void show(String[] strs);

    /**
     * Shows the command.
     * @param command
     */
    public synchronized void showCommand(String command, String[] args) {
        if (propagations != null) {
            try {
                String[] args1 = new String[propagations.size()];
                for (int i = 0; i < propagations.size(); i++) {
                    int lit = propagations.get(i);
                    decoder.propagate(lit);
                    args1[i] = decoder.decode(lit); 
                }
                propagations = null;
                showAssignment();
                showCommand("propagate", args1);
                decoder.saveState();
            } catch (TracerException e) {
                throw new IllegalArgumentException(e.getMessage());
            }
        }
        if (undos != null) {
            try {
                String[] args1 = new String[undos.size()];
                for (int i = 0; i < undos.size(); i++) {
                    int lit = undos.get(i);
                    decoder.undo(lit);
                    args1[i] = decoder.decode(lit); 
                }
                undos = null;
                showAssignment();
                showCommand("undo", args1);
                decoder.saveState();
            } catch (TracerException e) {
                throw new IllegalArgumentException(e.getMessage());
            }
        }
        String[] strs = null;
        if (args == null) {
            strs = new String[] { command };
        } else {
            strs = new String[args.length + 1];
            strs[0] = command;
            for (int i = 0; i < args.length; i++)
                strs[i + 1] = args[i];
        }
        show(strs);
    }
    
    /**
     * Shows the current assignment of the variable.
     * @param v
     */
    public synchronized void showAssignment(Variable v) {
        if ((showAuxVars || ! v.isAux()) && v.isModified(decoder.getAssignment()))
            showCommand("var", v.toStrings(decoder.getAssignment(), true));
    }

    /**
     * Shows the current assignment.
     */
    public synchronized void showAssignment() { 
        for (Variable v : decoder.getVariables())
            showAssignment(v);
    }
    
    /**
     * Starts the tracer.
     * @throws TracerException 
     * @see SatSolver#startSolver()
     */
    public synchronized void startTracer() throws TracerException {
        propagations = null;
        undos = null;
        done = false;
        decoder.reset();
        showAssignment();
        decoder.saveState();
        showCommand("startSolver", null);
        solver.startSolver();
    }
    
    /**
     * Stops the tracer.
     * @throws TracerException 
     * @see SatSolver#stopSolver()
     */
    public synchronized void stopTracer() throws TracerException {
        solver.stopSolver();
        showCommand("stopSolver", null);
    }
    
    /**
     * Called when decision is done for the literal by SAT solver.
     * @param lit the literal
     */
    public void decide(int lit) {
        try {
            decoder.decide(lit);
            showAssignment();
            decoder.saveState();
            showCommand("decide", new String[] { decoder.decode(lit) });
        } catch (TracerException e) {
            throw new IllegalArgumentException(e.getMessage());
        }
    }

    /**
     * Called when propagation is done for the literal by SAT solver.
     * @param lit the literal
     */
    public void propagate(int lit) {
        if (quick) {
            if (propagations == null)
                propagations = new ArrayList<Integer>();
            // propagations.add(lit);
            decoder.propagate(lit);
        } else {
            try {
                Variable v = decoder.propagate(lit);
                if (v != null) {
                    showAssignment(v);
                    showCommand("propagate", new String[] { decoder.decode(lit) });
                    v.saveDomain(decoder.getAssignment());
                }
            } catch (TracerException e) {
                throw new IllegalArgumentException(e.getMessage());
            }
        }
    }

    /**
     * Called when undo is done for the literal by SAT solver.
     * @param lit the literal
     */
    public void undo(int lit) {
        if (done)
            return;
        if (quick) {
            if (undos == null)
                undos = new ArrayList<Integer>();
            // undos.add(lit);
            decoder.undo(lit);
        } else {
            try {
                decoder.undo(lit);
                showAssignment();
                decoder.saveState();
                showCommand("undo", new String[] { decoder.decode(lit) });
            } catch (TracerException e) {
                throw new IllegalArgumentException(e.getMessage());
            }
        }
    }

    /**
     * Called when backjump is done by SAT solver.
     * @param backjumpLevel
     */
    public void backjump(int backjumpLevel) {
        decoder.backjump(backjumpLevel);
        showCommand("backjump", new String[] { Integer.toString(backjumpLevel) });
    }

    /**
     * Called when learnt clause is added by SAT solver.
     * The clauses is added to the learnt clauses.
     * @param clause the learnt clause
     */
    public void learnt(int[] clause) {
        try {
            decoder.learnt(clause);
            showCommand("learnt", decoder.decode(clause));
        } catch (TracerException e) {
            throw new IllegalArgumentException(e.getMessage());
        }
    }

    /**
     * Called when conflict happened at the clause by SAT solver.
     * @param clause the clause
     */
    public void conflict(int[] clause) {
        try {
            decoder.conflict(clause);
            showCommand("conflict", decoder.decode(clause));
        } catch (TracerException e) {
            throw new IllegalArgumentException(e.getMessage());
        }
    }

    /**
     * Called when restart happened by SAT solver.
     */
    public void restart() {
        try {
            showAssignment();
            decoder.saveState();
            showCommand("restart", null);
        } catch (TracerException e) {
            throw new IllegalArgumentException(e.getMessage());
        }
    }

    /**
     * Called when the solution was found by SAT solver.
     */
    public void found() {
        try {
            showAssignment();
            decoder.saveState();
            showCommand("found", null);
            done = true;
        } catch (TracerException e) {
            throw new IllegalArgumentException(e.getMessage());
        }
    }

    /**
     * Called when the result was returned from SAT solver.
     * @param result the result of SAT (true) or UNSAT (false)
     */
    public void end(boolean result) {
        try {
            decoder.end(result);
            showAssignment();
            decoder.saveState();
            done = true;
            if (result)
                showCommand("sat", null);
            else
                showCommand("unsat", null);
        } catch (TracerException e) {
            throw new IllegalArgumentException(e.getMessage());
        }
    }
}
