package sugartracer.decoder;

import java.util.BitSet;

/**
 * Maintains assignment on a SAT instance under solving. 
 * @author Naoyuki Tamura
 */
public class Assignment {
	private BitSet defined = new BitSet();
	private BitSet value = new BitSet();

    /**
     * Returns true when i-th propositional variable is defined. 
     * @param i index of the propositional variable to be checked
     * @return true when the propositional variable i is defined
     */
    public boolean isDefined(int i) {
        return defined.get(i);
    }
    
	/**
	 * Returns the Boolean value of i-th propositional variable.
	 * It returns null when the value is undefined. 
	 * @param i index of the propositional variable to be checked
	 * @return Boolean value of i-th propositional variable
	 */
	public Boolean get(int i) {
		if (! defined.get(i)) {
			return null;
		}
		return value.get(i) ? Boolean.TRUE : Boolean.FALSE;
	}
	
    /**
     * Sets the value of i-th propositional variable.
     * @param i index of the propositional variable to be set
     * @param val the value to be set
     */
	public void set(int i, boolean val) {
		defined.set(i);
		value.set(i, val);
	}

    /**
     * Sets i-th propositional variable undefined.
     * @param i index of the propositional variable to be unset
     */
    public void unset(int i) {
        defined.set(i, false);
        defined.clear(i);
    }
    
    /**
     * Returns true when the literal lit is defined. 
     * @param lit literal
     * @return true when lit is defined
     */
    public boolean isDefinedLit(int lit) {
        return defined.get(Math.abs(lit));
    }
    
    /**
     * Returns the Boolean value of the literal lit.
     * @param lit literal
     * @return the Boolean value
     */
    public Boolean getLit(int lit) {
        Boolean b = get(Math.abs(lit));
        if (b == null)
            return null;
        return lit > 0 ? b : ! b;
    }
    
    /**
     * Sets the value of the literal lit to be true.
     * @param lit the literal to be set
     */
    public void setLit(int lit) {
        set(Math.abs(lit), lit > 0);
    }
    
    /**
     * Sets the literal lit undefined.
     * @param lit the literal to be unset
     */
	public void unsetLit(int lit) {
		unset(Math.abs(lit));
	}
}
