package sugartracer.decoder;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Stack;

import sugartracer.TracerException;

/**
 * The class for decoding SAT solver information to CSP information.
 * @author Naoyuki Tamura
 */
public class Decoder {
    private EncodingMap encodingMap = null;
    private Assignment assignment = null;
    private Stack<Integer> decisions = null;
    private Stack<Integer> trail = null;
    private List<int[]> learnts = null;
    private Boolean result = null; 
    
    /**
     * Constructs the decoder using given encoding map
     * @param encodingMap the encoding map
     */
    public Decoder(EncodingMap encodingMap) {
        this.encodingMap = encodingMap;
        reset();
    }
    
    /**
     * Constructs the decoder using given map file
     * @param mapFileName map file name
     * @throws TracerException 
     */
    public Decoder(String mapFileName) throws TracerException {
        this(new EncodingMap());
        try {
            encodingMap.load(mapFileName);
        } catch (IOException e) {
            throw new TracerException(e.getMessage());
        }
    }

    /**
     * Resets the decoder to initial state 
     */
    public void reset() {
        assignment = new Assignment();
        decisions = new Stack<Integer>();
        trail = new Stack<Integer>();
        learnts = new ArrayList<int[]>();
        result = null;
        for (Variable v : getVariables())
            v.reset();
    }
    
	/**
	 * Returns the encoding map
	 * @return the encoding map
	 */
	public EncodingMap getEncodingMap() {
        return encodingMap;
    }

    /**
     * Returns the list of SAT variables
     * @return the list of SAT variables
     */
    public List<Variable> getVariables() {
        return encodingMap.getVariables();
    }
    
    /**
     * Returns the current assignment
     * @return the assignment
     */
    public Assignment getAssignment() {
        return assignment;
    }

    /**
     * Returns the current decision stack
     * @return the decision stack
     */
    public Stack<Integer> getDecisions() {
        return decisions;
    }

    /**
     * Returns the current trail stack
     * @return the trail stack
     */
    public Stack<Integer> getTrail() {
        return trail;
    }

    /**
     * Returns the current learnt clauses
     * @return the learnt clauses
     */
    public List<int[]> getLearnts() {
        return learnts;
    }
    
    /**
     * Returns the result of SAT solving.
     * @return the result
     */
    public Boolean getResult() {
        return result;
    }

    /**
     * Returns the string representation of the literal.
     * @param lit the literal
     * @return the string representation of the literal
     * @throws TracerException
     */
    public String decode(int lit) throws TracerException {
        boolean negative = lit < 0;
        int i = Math.abs(lit);
        Variable v = null;
        if (encodingMap != null)
            v = encodingMap.findVariable(i);
        if (v == null || v.isBool()) {
            String sign = negative ? "-" : "";
            String name = v == null ? Integer.toString(i) : v.getName();
            return sign + name;
        } else {
            String name = v.getName();
            String cmp = negative ? ">" : "<=";
            int value = v.toValue(Math.abs(i));
            return name + cmp + value;
        }
    }

    /**
     * Returns the list of string representations of literals
     * @param lits literals
     * @return the list of string representations
     * @throws TracerException
     */
    public List<String> decode(List<Integer> lits) throws TracerException {
        List<String> list = new ArrayList<String>();
        for (int lit : lits) {
            list.add(decode(lit));
        }
        return list;
    }
    
    /**
     * Returns the array of string representations of literals
     * @param lits literals
     * @return the array of string representations
     * @throws TracerException
     */
    public String[] decode(int[] lits) throws TracerException {
        String[] list = new String[lits.length];
        for (int i = 0; i < lits.length; i++)
            list[i] = decode(lits[i]);
        return list;
    }
    
    /**
     * Returns the list of string representations of clauses
     * @param clauses clauses
     * @return the list of string representations
     * @throws TracerException
     */
    public List<String[]> decodeClauses(List<int[]> clauses) throws TracerException {
        List<String[]> list = new ArrayList<String[]>();
        for (int[] lits : clauses) {
            list.add(decode(lits));
        }
        return list;
    }
    
    /**
     * Returns the string representation of the assignment.
     * @param includeAux true to include aux variables
     * @return the string representation of the assignment.
     */
    public List<String[]> getAssignmentDecoded(boolean includeAux) {
        List<String[]> list = new ArrayList<String[]>();
        for (Variable v : getVariables()) {
            if (includeAux || ! v.isAux()) {
                String[] data = new String[3];
                data[0] = v.toString();
                
            }
        }
        return list;
    }
    
    /**
     * Returns the string representation of decision stack.
     * @return the string representation of decision stack
     * @throws TracerException
     */
    public List<String> getDecisionsDecoded() throws TracerException {
        return decode(decisions);
    }

    /**
     * Returns the string representation of trail stack.
     * @return the string representation of trail stack
     * @throws TracerException
     */
    public List<String> getTrailDecoded() throws TracerException {
        return decode(trail);
    }

    /**
     * Returns the string representation of learnt clauses.
     * @return the string representation of learnt clauses
     * @throws TracerException
     */
    public List<String[]> getLearntsDecoded() throws TracerException {
        return decodeClauses(learnts);
    }

    /**
     * Saves the current assignment.
     * @throws TracerException
     */
    public void saveState() throws TracerException {
        for (Variable v : getVariables()) {
            v.saveDomain(assignment);
        }
    }
    
    /**
     * Called when decision is done for the literal by SAT solver.
     * The literal is added to decision stack.
     * @param lit the literal
     */
    public void decide(int lit) {
        decisions.push(lit);
    }

    /**
     * Called when propagation is done for the literal by SAT solver.
     * It returns the variable when its domain is updated.
     * The assignment is updated.
     * @param lit the literal
     * @return the variable when it is modified, otherwise null
     */
    public Variable propagate(int lit) {
        trail.push(lit);
        Variable v = encodingMap.findVariable(Math.abs(lit));
        if (v != null) {
            String oldDomain = v.domainString(assignment);
            assignment.setLit(lit);
            String newDomain = v.domainString(assignment);
            boolean modified = ! newDomain.equals(oldDomain);
            if (modified) {
                return v;
            }
        }
        return null;
    }

    /**
     * Called when undo is done for the literal by SAT solver.
     * It cancels assignments not only the literal but also literals propagated after the decision on lit.
     * @param lit the literal
     */
    public void undo(int lit) {
        while (trail.size() > 0) {
            int lit1 = trail.pop();
            assignment.unsetLit(lit1);
            if (lit == lit1)
                break;
        }
    }

    /**
     * Called when backjump is done by SAT solver.
     * The decision stack is truncated to the backjumpLevel.
     * @param backjumpLevel
     */
    public void backjump(int backjumpLevel) {
        while (decisions.size() > backjumpLevel)
            decisions.pop();
        // decisions = decisions.subList(0, backjumpLevel);
    }

    /**
     * Called when learnt clause is added by SAT solver.
     * The clauses is added to the learnt clauses.
     * @param clause the learnt clause
     */
    public void learnt(int[] clause) {
        learnts.add(clause);
    }

    /**
     * Called when conflict happened at the clause by SAT solver.
     * Currently it does nothing.
     * @param clause the clause
     */
    public void conflict(int[] clause) {
    }

    /**
     * Called when restart happened by SAT solver.
     * Currently it does nothing.
     */
    public void restart() {
    }

    /**
     * Called when the search was done by SAT solver.
     * @param result the result of SAT (true) or UNSAT (false)
     */
    public void end(boolean result) {
        this.result = result;
    }
}
