package sugartracer.decoder;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;

/**
 * Represents the mapping from CSP variables to SAT variables. 
 * @author Naoyuki Tamura
 */
public class EncodingMap {
    @SuppressWarnings("unused")
    private String objectiveTarget = null;
    @SuppressWarnings("unused")
    private String objectiveVariableName = null;
	private List<Variable> variables = null;
	private Map<String,Variable> nameVariableMap = null;
	private SortedMap<Integer,Variable> codeVariableMap = null;
	
	public EncodingMap() {
		variables = new ArrayList<Variable>();
		nameVariableMap = new HashMap<String,Variable>();
		codeVariableMap = new TreeMap<Integer,Variable>();
	}

    /**
     * Returns the size of CSP variables.
     * @return size of CSP variables
     */
    public int size() {
        return variables.size();
    }
    
    /**
     * Returns the list of CSP variables.
     * @return list of CSP variables
     */
	public List<Variable> getVariables() {
		return variables;
	}
	
    /**
     * Finds the CSP variable of the given name.
     * @param name the name of the variable
     * @return the CSP variable
     */
	public Variable findVariable(String name) {
		Variable v = nameVariableMap.get(name);
		return v;
	}
	
    /**
     * Finds the CSP variable whose encoded form uses the propositional variable code.
     * @param code index of the propositional variable 
     * @return the CSP variable
     */
	public Variable findVariable(int code) {
		code = Math.abs(code);
		Variable v = codeVariableMap.get(code);
		if (v == null) {
			SortedMap<Integer,Variable> head = codeVariableMap.headMap(code);
			int code0 = head.lastKey();
			v = codeVariableMap.get(code0);
			if (! v.contains(code)) {
				v = null;
			}
		}
		return v;
	}
	
    /**
     * Adds a CSP integer variable to this object. 
     * @param name
     * @param code
     * @param domain
     */
    public void addIntVariable(String name, int code, int[][] domain) {
        Variable v = Variable.newIntVariable(name, code, domain);
        variables.add(v);
        nameVariableMap.put(name, v);
        if (! (domain.length == 1 && domain[0][0] == domain[0][1]))
            codeVariableMap.put(code, v);
    }

    /**
     * Adds a CSP boolean variable to this object. 
     * @param name
     * @param code
     */
    public void addBoolVariable(String name, int code) {
        Variable v = Variable.newBoolVariable(name, code);
        variables.add(v);
        nameVariableMap.put(name, v);
        codeVariableMap.put(code, v);
    }
    
	/**
	 * Loads the mapping from the map file generated by Sugar.
	 * @param mapFileName the map file to be loaded
	 * @throws IOException
	 */
	public void load(String mapFileName) throws IOException {
		BufferedReader rd = new BufferedReader(
				new InputStreamReader(new FileInputStream(mapFileName), "UTF-8"));
		while (true) {
			String line = rd.readLine();
			if (line == null)
				break;
			String[] s = line.split("\\s+");
			if (s[0].equals("objective")) {
				objectiveTarget = s[1];
				objectiveVariableName = s[2];
			} else if (s[0].equals("int")) {
				String name = s[1];
				int code = Integer.parseInt(s[2]);
				int[][] domain = {};
				if (s.length == 4) {
					int lb;
					int ub;
					int pos = s[3].indexOf("..");
					if (pos < 0) {
						lb = ub = Integer.parseInt(s[3]);
					} else {
						lb = Integer.parseInt(s[3].substring(0, pos));
						ub = Integer.parseInt(s[3].substring(pos+2));
					}
					domain = new int[][] { { lb, ub } };
				} else {
					List<int[]> list = new ArrayList<int[]>();
					for (int i = 3; i < s.length; i++) {
						int lb;
						int ub;
						int pos = s[i].indexOf("..");
						if (pos < 0) {
							lb = ub = Integer.parseInt(s[i]);
						} else {
							lb = Integer.parseInt(s[i].substring(0, pos));
							ub = Integer.parseInt(s[i].substring(pos+2));
						}
						list.add(new int[] { lb, ub });
					}
					domain = list.toArray(domain);
				}
				addIntVariable(name, code, domain);
			} else if (s[0].equals("bool")) {
				String name = s[1];
				int code = Integer.parseInt(s[2]);
				addBoolVariable(name, code);
			}
		}
		rd.close();
	}

	private String toString(String prefix) {
		StringBuilder sb = new StringBuilder();
		for (Variable v : codeVariableMap.values()) {
			sb.append(prefix);
			sb.append(v.toString());
			sb.append("\n");
		}
		return sb.toString();
	}

	public String toString() {
	    return toString("");
	}

}
