package sugartracer.decoder;

import sugartracer.TracerException;

/**
 * This is a class for CSP integer variables and boolean variables. 
 * @author Naoyuki Tamura
 */
public class Variable {
	private String name;
	private int codeMin;
	private int codeMax;
	private int[][] domain;
    private int lbSaved = Integer.MIN_VALUE;
    private int ubSaved = Integer.MAX_VALUE;
	
	/**
	 * Creates a new CSP integer variable object.
     * @param name then name of the variable
     * @param code the minimum propositional variable index used to encode this variable
	 * @param domain array of intervals
	 * @return the CSP variable created
	 */
	public static Variable newIntVariable(String name, int code, int[][] domain) {
		return new Variable(name, code, domain);
	}
	
	/**
	 * Creates a new CSP boolean variable object.
     * @param name then name of the variable
     * @param code the minimum propositional variable index used to encode this variable
	 * @return the CSP variable created
	 */
	public static Variable newBoolVariable(String name, int code) {
		return new Variable(name, code, null);
	}
	
	private Variable(String name, int code, int[][] domain) {
		this.name = name;
		this.codeMin = code;
		this.domain = domain;
		int size = 0;
		if (domain == null) {
			// bool variable
			size = 2;
		} else {
			// int variable
			for (int[] range : domain) {
				size += range[1] - range[0] + 1;
			}
		}
		codeMax = codeMin + size - 2;
	}

	/**
	 * Resets the variable to initial state 
	 */
	public void reset() {
	    lbSaved = Integer.MIN_VALUE;
	    ubSaved = Integer.MAX_VALUE;
	}
	
	/**
	 * Returns the name of this variable.
	 * @return the name of this variable
	 */
	public String getName() {
	    return name;
	}
	
    /**
     * Returns true when this object is a CSP boolean variable.
     * @return true when this object is a CSP boolean variable
     */
	public boolean isBool() {
		return domain == null;
	}

    /**
     * Returns true when this object is an aux variable.
     * @return true when this object is an aux variable
     */
	public boolean isAux() {
		return name.startsWith("$");
	}

    /**
     * Returns true when the propositional variable code is used for this variable. 
     * @param code the index of the propositional variable
     * @return true when code is used for this
     */
	public boolean contains(int code) {
		return codeMin <= code && code <= codeMax;
	}
	
    /**
     * Returns the minimum propositional variable index used to encode this variable.
     * @return the minimum propositional variable index
     */
    public int getCodeMin() {
        return codeMin;
    }

    /**
     * Returns the maximum propositional variable index used to encode this variable.
     * @return the maximum propositional variable index
     */
    public int getCodeMax() {
        return codeMax;
    }

    /**
     * Returns the lower bound value of this variable.
     * 0 is returned when this is a boolean variable.
     * @return the lower bound value
     */
    public int lb() {
        return domain == null ? 0 : domain[0][0];
    }
    
    /**
     * Returns the upper bound value of this variable.
     * 1 is returned when this is a boolean variable.
     * @return the upper bound value
     */
    public int ub() {
        return domain == null ? 1 : domain[domain.length - 1][1];
    }
        
	/**
	 * Returns the value of this integer variable corresponding to code.
	 * Exception is raised when this is a boolean variable, or code is not used to encode this object.  
	 * @param code the index of the propositional variable
	 * @return the value of this integer variable corresponding
	 * @throws TracerException 
	 */
	public int toValue(int code) throws TracerException {
		if (isBool() || code < codeMin || codeMax < code)
			throw new TracerException("Internal error for " + code + " in " + this);
		int i = code - codeMin;
		for (int[] range : domain) {
			if (range[0] + i <= range[1]) {
				return range[0] + i;
			}
			i -= range[1] - range[0] + 1;
		}
		throw new TracerException("Internal error value(" + code + ") for " + this);
	}
	
	/**
	 * Returns the index of the propositional variable corresponding to the value.
	 * @param value the value of the variable
	 * @return the index of the propositional variable corresponding
	 * @throws TracerException
	 */
	public int toCode(int value) throws TracerException {
		if (isBool() || value < lb() || ub() <= value)
			throw new TracerException("Internal error for " + value + " in " + this);
		int code = codeMin;
		for (int[] range : domain) {
			if (range[0] <= value && value <= range[1]) {
				return code + value - range[0];
			}
			code += range[1] - range[0] + 1;
		}
		throw new TracerException("Internal error for " + value + " in " + this);
	}
	
    /**
     * Returns the lower bound value of this variable under given assignment.
     * @param assignment the assignment
     * @return the lower bound value under assignment
     */
	public int lb(Assignment assignment) {
        try {
	        if (assignment == null)
	            return lb();
	        if (isBool())
	            return assignment.get(codeMin) == Boolean.TRUE ? 1 : 0;
	        for (int code = codeMax; code >= codeMin; code--) {
	            if (assignment.get(code) == Boolean.FALSE)
	                return code == codeMax ? ub() : toValue(code + 1);
	        }
        } catch (TracerException e) {
        }
		return lb();
	}
	
    /**
     * Returns the upper bound value of this variable under given assignment.
     * @param assignment the assignment
     * @return the upper bound value under assignment
     */
	public int ub(Assignment assignment) {
        try {
            if (assignment == null)
                return ub();
            if (isBool())
                return assignment.get(codeMin) == Boolean.FALSE ? 0 : 1;
            for (int code = codeMin; code <= codeMax; code++) {
                if (assignment.get(code) == Boolean.TRUE)
                    return toValue(code);
            }
        } catch (TracerException e) {
        }
		return ub();
	}

    /**
     * Saves the domain information under given assignment.
     * @param assignment the assignment
     */
    public void saveDomain(Assignment assignment) {
        lbSaved = lb(assignment);
        ubSaved = ub(assignment);
    }

    /**
     * Returns true when the domain is modified.
     * @param assignment
     * @return true when the domain is modified
     */
    public boolean isModified(Assignment assignment) {
        return lbSaved != lb(assignment) || ubSaved != ub(assignment);
    }
    
    private String domainString(int lb, int ub) {
		StringBuilder sb = new StringBuilder();
		if (isBool()) {
			sb.append(lb < ub ? "undef" : (lb == 0 ? "false" : "true"));
		} else {
			String delim = "";
			for (int[] range : domain) {
				int lb0 = Math.max(lb, range[0]);
				int ub0 = Math.min(ub, range[1]);
				if (lb0 <= ub0) {
					sb.append(delim);
					sb.append(lb0);
					if (lb0 < ub0) {
						sb.append("..");
						sb.append(ub0);
					}
					delim = ",";
				}
			}
			if (sb.length() == 0) {
				sb.append("empty");
			}
		}
		return sb.toString();
	}
	
	/**
     * Returns the string representation of the domain under given assignment  
	 * @param assignment the assignment
	 * @return the string representation
	 */
	public String domainString(Assignment assignment) {
	    return domainString(lb(assignment), ub(assignment));
	}
	
	/**
     * Returns the string representation of the saved domain
	 * @return the string representation
	 */
	public String domainStringSaved() {
	    if (lbSaved == Integer.MIN_VALUE && ubSaved == Integer.MAX_VALUE)
	        return "undef";
	    return domainString(lbSaved, ubSaved);
	}
	
	/**
	 * Returns the string representation of this variable
     * @param assignment
	 * @param showOldDomain
	 * @return the string representation of this variable
	 */
	public String[] toStrings(Assignment assignment, boolean showOldDomain) {
        String type = isBool() ? "bool" : "int";
	    if (showOldDomain && isModified(assignment)) {
	        return new String[] { type, name, domainString(assignment), domainStringSaved() }; 
	    } else {
	        return new String[] { type, name, domainString(assignment) }; 
	    }
	}

	public String toString() {
        String type = isBool() ? "bool" : "int";
		return type + " " + name + " " + domainString(null);
	}
}
